@@ -8,6 +8,7 @@ from asn1crypto.core import VOID
 from cryptography.hazmat.primitives import hashes, hmac, keywrap
 from cryptography.hazmat.primitives.kdf import hkdf
 from pyhanko.pdf_utils import generic, misc
+from pyhanko_certvalidator.util import get_pyca_cryptography_hash
 
 from ...sign.general import (
     CMSStructuralError,
@@ -16,7 +17,6 @@ from ...sign.general import (
     ValueErrorWithMessage,
     byte_range_digest,
     find_unique_cms_attribute,
-    get_pyca_cryptography_hash,
     simple_cms_attribute,
 )
 from ...sign.signers.pdf_byterange import (
@@ -36,11 +36,11 @@ from ..writer import BasePdfFileWriter
 from ._iso32004_asn1 import PdfMacIntegrityInfo
 
 __all__ = [
-    'PdfMacTokenHandler',
-    'validate_pdf_mac',
-    'add_standalone_mac',
-    'ISO32004',
     'ALLOWED_MD_ALGS',
+    'ISO32004',
+    'PdfMacTokenHandler',
+    'add_standalone_mac',
+    'validate_pdf_mac',
 ]
 
 
@@ -63,6 +63,7 @@ ALLOWED_MD_ALGS = frozenset(
         'sha3_384',
         'sha512',
         'sha3_512',
+        'shake256',
     ]
 )
 """
