@@ -37,32 +37,33 @@ if typing.TYPE_CHECKING:
     from .crypt.api import SecurityHandler
 
 __all__ = [
-    'Dereferenceable',
-    'Reference',
-    'TrailerReference',
-    'PdfObject',
-    'IndirectObject',
-    'NullObject',
-    'BooleanObject',
-    'FloatObject',
-    'NumberObject',
-    'ByteStringObject',
-    'TextStringObject',
-    'NameObject',
     'ArrayObject',
+    'BooleanObject',
+    'ByteStringObject',
+    'DecryptedObjectProxy',
+    'Dereferenceable',
     'DictionaryObject',
+    'EncryptedObjAccess',
+    'FloatObject',
+    'IndirectObject',
+    'NameObject',
+    'NullObject',
+    'NumberObject',
+    'PdfObject',
+    'Reference',
     'StreamObject',
-    'read_object',
+    'TextStringEncoding',
+    'TextStringObject',
+    'TrailerReference',
+    'pdf_date',
     'pdf_name',
     'pdf_string',
-    'pdf_date',
-    'TextStringEncoding',
-    'EncryptedObjAccess',
-    'DecryptedObjectProxy',
+    'read_object',
+    'OperatorLiteral',
 ]
 
 OBJECT_PREFIXES = b'/<[tf(n%'
-NUMBER_SIGNS = b'+-'
+NUMBER_SIGNS = b'+-.'
 INDIRECT_PATTERN = re.compile(r"(\d+)\s+(\d+)\s+R[^a-zA-Z]".encode('ascii'))
 
 logger = logging.getLogger(__name__)
@@ -207,8 +208,27 @@ class Reference(Dereferenceable):
         return self.pdf
 
 
+@enum.unique
+class ContentOpReading(enum.Enum):
+    NO = 1
+    GRAPHIC_OPS = 2
+    IN_OBJECT = 3
+
+    def __bool__(self):
+        return self != ContentOpReading.NO
+
+    def nest(self):
+        if not self:
+            return ContentOpReading.NO
+        else:
+            return ContentOpReading.IN_OBJECT
+
+
 def read_object(
-    stream, container_ref: 'Dereferenceable', as_metadata_stream: bool = False
+    stream,
+    container_ref: 'Dereferenceable',
+    as_metadata_stream: bool = False,
+    in_content_stream: ContentOpReading = ContentOpReading.NO,
 ) -> 'PdfObject':
     """
     Read a PDF object from an input stream.
@@ -228,6 +248,9 @@ def read_object(
         resolve to the return value of this function.
     :param as_metadata_stream:
         Whether to dereference the object as an XMP metadata stream.
+    :param in_content_stream:
+        Whether we are parsing inside a content stream (where
+        certain kinds of object cannot occur).
     :return:
         A :class:`.PdfObject`.
     """
@@ -244,22 +267,24 @@ def read_object(
         stream.seek(-2, os.SEEK_CUR)  # reset to start
         if peek == b'<<':
             result = DictionaryObject.read_from_stream(
-                stream, container_ref, as_metadata_stream=as_metadata_stream
+                stream,
+                container_ref,
+                as_metadata_stream=as_metadata_stream,
+                in_content_stream=in_content_stream.nest(),
             )
         else:
             result = read_hex_string_from_stream(stream)
     elif idx == 2:
         # array object
-        result = ArrayObject.read_from_stream(stream, container_ref)
+        result = ArrayObject.read_from_stream(
+            stream, container_ref, in_content_stream=in_content_stream.nest()
+        )
     elif idx == 3 or idx == 4:
         # boolean object
         result = BooleanObject.read_from_stream(stream)
     elif idx == 5:
         # string object
         result = read_string_from_stream(stream)
-    elif idx == 6:
-        # null object
-        result = NullObject.read_from_stream(stream)
     elif idx == 7:
         # comment
         while tok not in (b'\r', b'\n'):
@@ -267,18 +292,38 @@ def read_object(
         read_non_whitespace(stream)
         stream.seek(-1, os.SEEK_CUR)
         result = read_object(stream, container_ref)
-    else:
-        # number object OR indirect reference
-        if tok in NUMBER_SIGNS:
-            # number
+    elif in_content_stream == ContentOpReading.GRAPHIC_OPS and tok.isalpha():
+        # operator literal
+        literal = read_until_delimiter(stream)
+        if not literal.isascii():
+            raise PdfReadError(
+                f"Failed to parse {literal!r} as operator literal"
+            )
+        elif literal == b'null':
+            return NullObject()
+        elif literal == b"R":
+            raise PdfReadError("Indirect object reference in content stream")
+        result = OperatorLiteral(literal.decode('ascii'))
+    elif idx == 6:
+        # null object
+        result = NullObject.read_from_stream(stream)
+    elif tok in NUMBER_SIGNS:
+        result = NumberObject.read_from_stream(stream)
+    elif tok.isdigit():
+        if in_content_stream:
             result = NumberObject.read_from_stream(stream)
         else:
+            # number object OR indirect reference
             peek = stream.read(20)
             stream.seek(-len(peek), os.SEEK_CUR)  # reset to start
             if INDIRECT_PATTERN.match(peek) is not None:
                 result = IndirectObject.read_from_stream(stream, container_ref)
             else:
                 result = NumberObject.read_from_stream(stream)
+    else:
+        raise PdfReadError(
+            f"Unexpected byte '{tok}' at position {stream.tell()}"
+        )
 
     result.container_ref = container_ref
     return result
@@ -353,6 +398,19 @@ class PdfObject:
         raise NotImplementedError
 
 
+class OperatorLiteral(PdfObject):
+    def __init__(self, literal: str):
+        self.literal = literal
+
+    def __eq__(self, other):
+        return (
+            isinstance(other, OperatorLiteral) and self.literal == other.literal
+        )
+
+    def __hash__(self):
+        return hash(("op", self.literal))
+
+
 class NullObject(PdfObject):
     """
     PDF `null` object.
@@ -448,6 +506,12 @@ class ArrayObject(list, PdfObject):
     def __getitem__(self, index):
         return self.raw_get(index).get_object()
 
+    def iterate(
+        self, decrypt: EncryptedObjAccess = EncryptedObjAccess.TRANSPARENT
+    ):
+        for ix in range(len(self)):
+            yield self.raw_get(ix, decrypt)
+
     def raw_get(
         self,
         index,
@@ -490,7 +554,11 @@ class ArrayObject(list, PdfObject):
         stream.write(b" ]")
 
     @staticmethod
-    def read_from_stream(stream, container_ref):
+    def read_from_stream(
+        stream,
+        container_ref,
+        in_content_stream: ContentOpReading = ContentOpReading.NO,
+    ):
         arr = ArrayObject()
         tmp = stream.read(1)
         if tmp != b"[":
@@ -502,7 +570,11 @@ class ArrayObject(list, PdfObject):
                 break
             stream.seek(-1, os.SEEK_CUR)
             # read and append obj
-            arr.append(read_object(stream, container_ref))
+            arr.append(
+                read_object(
+                    stream, container_ref, in_content_stream=in_content_stream
+                )
+            )
         return arr
 
 
@@ -1048,7 +1120,7 @@ def _decode_name(name_bytes: bytes) -> 'NameObject':
                     digit2 = next(name_iter)
                 except StopIteration:
                     raise PdfReadError(
-                        f"Unterminated escape in PDF name /{repr(name_bytes)}"
+                        f"Unterminated escape in PDF name /{name_bytes!r}"
                     )
 
                 cur_byte = _as_hex_digit(digit1) * 16 + _as_hex_digit(digit2)
@@ -1100,8 +1172,7 @@ class NameObject(str, PdfObject):
         byte_iter = iter(self.encode('utf8'))
         if not next(byte_iter) == 0x2F:
             raise PdfWriteError(
-                f"Could not serialise name object {repr(self)}, "
-                f"must start with /"
+                f"Could not serialise name object {self!r}, must start with /"
             )
         stream.write(b'/')
         for cur_byte in byte_iter:
@@ -1134,6 +1205,47 @@ def _normalise_key(key):
     return key
 
 
+def _maybe_read_stream_payload(stream, data, handler):
+    pos = stream.tell()
+    s = read_non_whitespace(stream, allow_eof=True)
+    stream_data = None
+    if s == b's' and stream.read(5) == b'tream':
+        # odd PDF file output has spaces after 'stream' keyword
+        # but before EOL. Original PyPDF2 patch provided by Danial Sandler,
+        # modified by Matthias Valvekens
+        skip_over_whitespace(stream, stop_after_eol=True)
+        # this is a stream object, not a dictionary
+        length = data[pdf_name("/Length")]
+        if isinstance(length, IndirectObject):
+            t = stream.tell()
+            length = handler.get_object(length)
+            stream.seek(t)
+        stream_data = stream.read(length)
+        e = read_non_whitespace(stream)
+        ndstream = stream.read(8)
+        if (e + ndstream) != b"endstream":
+            # (sigh) - the odd PDF file has a length that is too long, so
+            # we need to read backwards to find the "endstream" ending.
+            # ReportLab (unknown version) generates files with this bug,
+            # and Python users into PDF files tend to be our audience.
+            # we need to do this to correct the streamdata and chop off
+            # an extra character.
+            orig_endstream_pos = stream.tell()
+            stream.seek(-10, os.SEEK_CUR)
+            end = stream.read(9)
+            if end == b"endstream":
+                # we found it by looking back one character further.
+                stream_data = stream_data[:-1]
+            else:
+                raise PdfReadError(
+                    "Unable to find 'endstream' marker after "
+                    "stream at byte %s." % hex(orig_endstream_pos)
+                )
+    else:
+        stream.seek(pos)
+    return stream_data
+
+
 class DictionaryObject(dict, PdfObject):
     """
     A PDF dictionary object.
@@ -1255,6 +1367,7 @@ class DictionaryObject(dict, PdfObject):
         stream,
         container_ref: 'Dereferenceable',
         as_metadata_stream: bool = False,
+        in_content_stream: ContentOpReading = ContentOpReading.NO,
     ):
         tmp = stream.read(2)
         if tmp != b"<<":
@@ -1279,7 +1392,9 @@ class DictionaryObject(dict, PdfObject):
                 ) from ex
             read_non_whitespace(stream)
             stream.seek(-1, os.SEEK_CUR)
-            value = read_object(stream, container_ref)
+            value = read_object(
+                stream, container_ref, in_content_stream=in_content_stream
+            )
             if key not in data:
                 data[key] = value
             else:
@@ -1292,43 +1407,10 @@ class DictionaryObject(dict, PdfObject):
                 else:
                     logger.warning(err)
 
-        pos = stream.tell()
-        s = read_non_whitespace(stream, allow_eof=True)
-        stream_data = None
-        if s == b's' and stream.read(5) == b'tream':
-            # odd PDF file output has spaces after 'stream' keyword
-            # but before EOL. Original PyPDF2 patch provided by Danial Sandler,
-            # modified by Matthias Valvekens
-            skip_over_whitespace(stream, stop_after_eol=True)
-            # this is a stream object, not a dictionary
-            length = data[pdf_name("/Length")]
-            if isinstance(length, IndirectObject):
-                t = stream.tell()
-                length = handler.get_object(length)
-                stream.seek(t)
-            stream_data = stream.read(length)
-            e = read_non_whitespace(stream)
-            ndstream = stream.read(8)
-            if (e + ndstream) != b"endstream":
-                # (sigh) - the odd PDF file has a length that is too long, so
-                # we need to read backwards to find the "endstream" ending.
-                # ReportLab (unknown version) generates files with this bug,
-                # and Python users into PDF files tend to be our audience.
-                # we need to do this to correct the streamdata and chop off
-                # an extra character.
-                orig_endstream_pos = stream.tell()
-                stream.seek(-10, os.SEEK_CUR)
-                end = stream.read(9)
-                if end == b"endstream":
-                    # we found it by looking back one character further.
-                    stream_data = stream_data[:-1]
-                else:
-                    raise PdfReadError(
-                        "Unable to find 'endstream' marker after "
-                        "stream at byte %s." % hex(orig_endstream_pos)
-                    )
+        if in_content_stream:
+            stream_data = None
         else:
-            stream.seek(pos)
+            stream_data = _maybe_read_stream_payload(stream, data, handler)
         if stream_data is not None:
             # pass in everything as encoded data, the StreamObject class
             # will take care of decoding as necessary
@@ -1634,7 +1716,7 @@ class StreamObject(DictionaryObject):
 
             # prepend the new filter (order is important!)
             self[pdf_name('/Filter')] = ArrayObject(
-                (filter_name,) + filter_names
+                (filter_name, *filter_names)
             )
 
             if params or any(param_sets):
