@@ -11,11 +11,11 @@ from asn1crypto import x509
 from asn1crypto.x509 import KeyUsage
 from pyhanko.pdf_utils import generic
 from pyhanko.pdf_utils.content import RawContent
+from pyhanko.pdf_utils.form_tools import enumerate_fields_in
 from pyhanko.pdf_utils.generic import pdf_name, pdf_string
 from pyhanko.pdf_utils.layout import BoxConstraints
 from pyhanko.pdf_utils.misc import (
     OrderedEnum,
-    PdfError,
     PdfReadError,
     PdfWriteError,
     get_and_apply,
@@ -24,34 +24,31 @@ from pyhanko.pdf_utils.misc import (
 from pyhanko.pdf_utils.rw_common import PdfHandler
 from pyhanko.pdf_utils.writer import BasePdfFileWriter
 from pyhanko.sign.general import SigningError, UnacceptableSignerError
-
 from pyhanko_certvalidator.authority import AuthorityWithCert
 from pyhanko_certvalidator.errors import InvalidCertificateError
 from pyhanko_certvalidator.path import ValidationPath
 
 __all__ = [
-    'SigFieldSpec',
-    'SigSeedValFlags',
-    'SigCertConstraints',
-    'SigSeedValueSpec',
-    'SigCertConstraintFlags',
-    'SigSeedSubFilter',
-    'SeedValueDictVersion',
-    'SeedLockDocument',
-    'SigCertKeyUsage',
-    'MDPPerm',
     'FieldMDPAction',
     'FieldMDPSpec',
-    'SignatureFormField',
     'InvisSigSettings',
+    'MDPPerm',
+    'SeedLockDocument',
+    'SeedValueDictVersion',
+    'SigCertConstraintFlags',
+    'SigCertConstraints',
+    'SigCertKeyUsage',
+    'SigFieldSpec',
+    'SigSeedSubFilter',
+    'SigSeedValFlags',
+    'SigSeedValueSpec',
+    'SignatureFormField',
     'VisibleSigSettings',
-    'enumerate_sig_fields',
     'append_signature_field',
-    'ensure_sig_flags',
-    'prepare_sig_field',
     'apply_sig_field_spec_properties',
-    'annot_width_height',
-    'get_sig_field_annot',
+    'ensure_sig_flags',
+    'enumerate_sig_fields',
+    'prepare_sig_field',
 ]
 
 logger = logging.getLogger(__name__)
@@ -1479,8 +1476,11 @@ def prepare_sig_field(
         except KeyError:
             fields = form['/Fields'] = generic.ArrayObject()
 
-        candidates = enumerate_sig_fields_in(
-            fields, with_name=sig_field_name, refs_seen=set()
+        candidates = enumerate_fields_in(
+            fields,
+            with_name=sig_field_name,
+            refs_seen=set(),
+            target_field_type='/Sig',
         )
         sig_field_ref = None
         try:
@@ -1534,51 +1534,6 @@ def prepare_sig_field(
     return True, sig_field_ref
 
 
-def get_sig_field_annot(
-    sig_field: generic.DictionaryObject,
-) -> generic.DictionaryObject:
-    """
-    Internal function to get the annotation of a signature field.
-
-    :param sig_field:
-        A signature field dictionary.
-    :return:
-        The dictionary of the corresponding annotation.
-    """
-    try:
-        (sig_annot,) = sig_field['/Kids']
-        sig_annot = sig_annot.get_object()
-    except (ValueError, TypeError):
-        raise SigningError(
-            "Failed to access signature field's annotation. "
-            "Signature field must have exactly one child annotation, "
-            "or it must be combined with its annotation."
-        )
-    except KeyError:
-        sig_annot = sig_field
-    return sig_annot
-
-
-def annot_width_height(
-    annot_dict: generic.DictionaryObject,
-) -> Tuple[float, float]:
-    """
-    Internal function to compute the width and height of an annotation.
-
-    :param annot_dict:
-        Annotation dictionary.
-    :return:
-        a (width, height) tuple
-    """
-    try:
-        x1, y1, x2, y2 = annot_dict['/Rect']
-    except KeyError:
-        return 0, 0
-    w = abs(x1 - x2)
-    h = abs(y1 - y2)
-    return w, h
-
-
 def enumerate_sig_fields(
     handler: PdfHandler,
     filled_status: Optional[bool] = None,
@@ -1604,102 +1559,15 @@ def enumerate_sig_fields(
     except KeyError:
         return
 
-    yield from enumerate_sig_fields_in(
+    yield from enumerate_fields_in(
         fields,
         filled_status=filled_status,
         with_name=with_name,
         refs_seen=set(),
+        target_field_type='/Sig',
     )
 
 
-def enumerate_sig_fields_in(
-    field_list,
-    filled_status=None,
-    with_name=None,
-    parent_name="",
-    parents=None,
-    *,
-    refs_seen,
-):
-    if not isinstance(field_list, generic.ArrayObject):
-        logger.warning(
-            f"Values of type {type(field_list)} are not valid as field "
-            f"lists, must be array objects -- skipping."
-        )
-        return
-
-    parents = parents or ()
-    for field_ref in field_list:
-        if not isinstance(field_ref, generic.IndirectObject):
-            logger.warning(
-                "Entries in field list must be indirect references -- skipping."
-            )
-            continue
-        if field_ref.reference in refs_seen:
-            raise PdfReadError("Circular reference in form tree")
-
-        field = field_ref.get_object()
-        if not isinstance(field, generic.DictionaryObject):
-            logger.warning(
-                "Entries in field list must be dictionary objects, not "
-                f"{type(field)} -- skipping."
-            )
-            continue
-        # /T is the field name. If not specified, we're dealing with a bare
-        # widget, so skip it. (these should never occur in /Fields, but hey)
-        try:
-            field_name = field['/T']
-        except KeyError:
-            continue
-        fq_name = (
-            field_name
-            if not parent_name
-            else ("%s.%s" % (parent_name, field_name))
-        )
-        explicitly_requested = with_name is not None and fq_name == with_name
-        child_requested = explicitly_requested or (
-            with_name is not None and with_name.startswith(fq_name)
-        )
-        # /FT is inheritable, so go up the chain
-        current_path = (field,) + parents
-        for parent_field in current_path:
-            try:
-                field_type = parent_field['/FT']
-                break
-            except KeyError:
-                continue
-        else:
-            field_type = None
-
-        if field_type == '/Sig':
-            field_value = field.get('/V')
-            # "cast" to a regular string object
-            filled = field_value is not None
-            status_check = filled_status is None or filled == filled_status
-            name_check = with_name is None or explicitly_requested
-            if status_check and name_check:
-                yield fq_name, field_value, field_ref
-        elif explicitly_requested:
-            raise SigningError(
-                'Field with name %s exists but is not a signature field'
-                % fq_name
-            )
-
-        # if necessary, descend into the field hierarchy
-        if with_name is None or (child_requested and not explicitly_requested):
-            try:
-                yield from enumerate_sig_fields_in(
-                    field['/Kids'],
-                    parent_name=fq_name,
-                    parents=current_path,
-                    with_name=with_name,
-                    filled_status=filled_status,
-                    refs_seen=refs_seen | {field_ref.reference},
-                )
-            except KeyError:
-                continue
-
-
 def append_signature_field(
     pdf_out: BasePdfFileWriter, sig_field_spec: SigFieldSpec
 ):
