@@ -7,7 +7,6 @@ seamlessly plugged into a :class:`~.signers.PdfSigner`.
 import asyncio
 import binascii
 import logging
-import struct
 import warnings
 from dataclasses import dataclass
 from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
@@ -22,10 +21,10 @@ from pyhanko.config.pkcs11 import (
     TokenCriteria,
 )
 from pyhanko.pdf_utils.misc import coalesce
-from pyhanko.sign.general import SigningError, get_pyca_cryptography_hash
+from pyhanko.sign.general import SigningError
 from pyhanko.sign.signers import Signer
-
 from pyhanko_certvalidator.registry import CertificateStore
+from pyhanko_certvalidator.util import get_pyca_cryptography_hash
 
 try:
     from pkcs11 import (
@@ -49,9 +48,9 @@ except ImportError as e:  # pragma: nocover
 
 __all__ = [
     'PKCS11Signer',
-    'open_pkcs11_session',
     'PKCS11SigningContext',
     'find_token',
+    'open_pkcs11_session',
     'select_pkcs11_signing_params',
 ]
 
@@ -107,12 +106,14 @@ def find_token(
                 "must be provided"
             )
 
+    token = None
     if slot_no is None:
         for slot in slots:
             try:
-                token = slot.get_token()
-                if criteria_satisfied_by(token_criteria, token):
-                    return token
+                cur_token = slot.get_token()
+                if criteria_satisfied_by(token_criteria, cur_token):
+                    token = cur_token
+                    break
             except PKCS11Error:
                 continue
     else:
@@ -120,8 +121,8 @@ def find_token(
             raise PKCS11Error(
                 f"Slot index {slot_no} too large; there are only {len(slots)}"
             )
-        token = slots[slot_no].get_token()
-        errors = criteria_mismatches(token_criteria, token)
+        cur_token = slots[slot_no].get_token()
+        errors = criteria_mismatches(token_criteria, cur_token)
         if errors:
             err_str = ", ".join(
                 f"{field} is not {val!r}" for field, val in errors
@@ -129,8 +130,9 @@ def find_token(
             raise PKCS11Error(
                 f"Token in slot {slot_no} does not satisfy criteria; {err_str}."
             )
-        return token
-    return None
+        else:
+            token = cur_token
+    return token
 
 
 @dataclass(frozen=True)
@@ -213,6 +215,7 @@ def select_pkcs11_signing_params(
     signature_mechanism: algos.SignedDigestAlgorithm,
     digest_algorithm: str,
     use_raw_mechanism: bool,
+    sign_kwargs: Dict[str, Any],
 ) -> PKCS11SignatureOperationSpec:
     """
     Internal helper function to set up a PKCS #11 signing operation.
@@ -223,6 +226,8 @@ def select_pkcs11_signing_params(
         The digest algorithm to use
     :param use_raw_mechanism:
         Whether to attempt to use the raw mechanism on pre-hashed data.
+    :param sign_kwargs:
+        Override kwargs manually.
     :return:
     """
     from pkcs11.util.dsa import encode_dsa_signature
@@ -294,22 +299,12 @@ def select_pkcs11_signing_params(
             # Note: Ed448-ph isn't the same thing.
             raise NotImplementedError("Ed448 not available in raw mode")
         kwargs['mechanism'] = Mechanism.EDDSA
-        # Definition of the param type:
-        # typedef struct CK_EDDSA_PARAMS {
-        #    CK_BBOOL     phFlag;
-        #    CK_ULONG     ulContextDataLen;
-        #    CK_BYTE_PTR  pContextData;
-        # }  CK_EDDSA_PARAMS;
-        # We use native size and alignment here on purpose
-
-        # NOTE: I _think_ this is correct, but it looks like SoftHSMv2
-        # doesn't really care about the params, so maybe I'm wrong.
-        kwargs['mechanism_param'] = struct.pack('@?LP', False, 0, 0)
-    else:
+        kwargs['mechanism_param'] = (False, None)
+    elif 'mechanism' not in sign_kwargs:
         raise NotImplementedError(
             f"Signature algorithm '{signature_algo}' is not supported."
         )
-
+    kwargs.update(**sign_kwargs)
     return PKCS11SignatureOperationSpec(
         sign_kwargs=kwargs,
         pre_sign_transform=pre_sign_transform,
@@ -377,48 +372,24 @@ def open_pkcs11_session(
 
 
 def _format_pull_err_msg(
+    object_kind: str,
     no_results: bool,
     label: Optional[str] = None,
-    cert_id: Optional[bytes] = None,
+    id_value: Optional[bytes] = None,
 ):
     info_strs = []
     if label is not None:
         info_strs.append(f"label '{label}'")
-    if cert_id is not None:
-        info_strs.append(f"ID '{binascii.hexlify(cert_id).decode('ascii')}'")
+    if id_value is not None:
+        info_strs.append(f"ID '{binascii.hexlify(id_value).decode('ascii')}'")
     qualifier = f" with {', '.join(info_strs)}" if info_strs else ""
     if no_results:
-        err = f"Could not find cert{qualifier}."
+        err = f"Could not find {object_kind}{qualifier}."
     else:
-        err = f"Found more than one cert{qualifier}."
+        err = f"Found more than one {object_kind}{qualifier}."
     return err
 
 
-def _pull_cert(
-    pkcs11_session: Session,
-    label: Optional[str] = None,
-    cert_id: Optional[bytes] = None,
-):
-    query_params = {Attribute.CLASS: ObjectClass.CERTIFICATE}
-    if label is not None:
-        query_params[Attribute.LABEL] = label
-    if cert_id is not None:
-        query_params[Attribute.ID] = cert_id
-    q = pkcs11_session.get_objects(query_params)
-
-    # need to run through the full iterator to make sure the operation
-    # terminates
-    results = list(q)
-    if len(results) == 1:
-        cert_obj = results[0]
-        return x509.Certificate.load(cert_obj[Attribute.VALUE])
-    else:
-        err = _format_pull_err_msg(
-            no_results=not results, label=label, cert_id=cert_id
-        )
-        raise PKCS11Error(err)
-
-
 def _hash_fully(digest_algorithm: str, *, wrap_digest_info: bool):
     md_spec = get_pyca_cryptography_hash(digest_algorithm)
 
@@ -492,8 +463,21 @@ class PKCS11Signer(Signer):
             This functionality is only available for ECDSA at this time.
             Support for other signature schemes will be added on an as-needed
             basis.
+    :param signature_mechanism:
+        Specify the signature mechanism explicitly. By default, pyHanko will
+        attempt to infer it from the key's properties.
+    :param base_sign_kwargs:
+        Directly specify some of the keyword arguments passed to the
+        :meth:`pkcs11.SignMixin.sign` method.
+
+        This can be used as an escape hatch to enable functionality
+        (advanced auth, additional signing mechanisms) that are not
+        directly supported by pyHanko.
     """
 
+    default_cert_query_params: Dict[Attribute, Any]
+    default_key_query_params: Dict[Attribute, Any]
+
     def __init__(
         self,
         pkcs11_session: Session,
@@ -509,6 +493,7 @@ class PKCS11Signer(Signer):
         cert_id: Optional[bytes] = None,
         use_raw_mechanism=False,
         signature_mechanism: Optional[algos.SignedDigestAlgorithm] = None,
+        base_sign_kwargs: Optional[Dict[str, Any]] = None,
     ):
         """
         Initialise a PKCS11 signer.
@@ -529,6 +514,7 @@ class PKCS11Signer(Signer):
         self.use_raw_mechanism = use_raw_mechanism
         self._key_handle = None
         self._loaded = False
+        self._sign_kwargs = base_sign_kwargs or {}
         self.__loading_event = None
         super().__init__(
             prefer_pss=prefer_pss,
@@ -540,6 +526,13 @@ class PKCS11Signer(Signer):
             self._cert_registry.register_multiple(ca_chain)
         if signing_cert is not None:
             self._cert_registry.register(signing_cert)
+        self.default_cert_query_params = {
+            Attribute.CLASS: ObjectClass.CERTIFICATE
+        }
+        self.default_key_query_params = {
+            Attribute.CLASS: ObjectClass.PRIVATE_KEY,
+            Attribute.SIGN: True,
+        }
 
     def _init_cert_registry(self):
         # it's conceivable that one might want to load this separately from
@@ -561,14 +554,39 @@ class PKCS11Signer(Signer):
         self._load_objects()
         return self._signing_cert
 
+    # noinspection PyUnusedLocal
+    def sign_kwargs(self, data_to_sign: bytes) -> Dict[str, Any]:
+        """
+        Supply keyword arguments to pass to the :meth:`pkcs11.SignMixin.sign`
+        method.
+
+        This can be used as an escape hatch to enable functionality
+        (advanced auth, additional signing mechanisms) that are not
+        directly supported by pyHanko.
+
+        :param data_to_sign:
+            The data being signed.
+
+            .. note::
+                In typical cases this parameter can be ignored;
+                pyHanko exposes it to subclasses to allow for
+                just-in-time credentials that depend on
+                the content being signed.
+
+        :return:
+            A dictionary of keyword arguments.
+        """
+        return self._sign_kwargs
+
     def _select_pkcs11_signing_params(
-        self, digest_algorithm: str
+        self, digest_algorithm: str, sign_kwargs: Dict[str, Any]
     ) -> PKCS11SignatureOperationSpec:
         digest_algorithm = digest_algorithm.lower()
         return select_pkcs11_signing_params(
             self.get_signature_mechanism_for_digest(digest_algorithm),
             digest_algorithm,
             use_raw_mechanism=self.use_raw_mechanism,
+            sign_kwargs=sign_kwargs,
         )
 
     async def async_sign_raw(
@@ -582,7 +600,9 @@ class PKCS11Signer(Signer):
         from pkcs11 import SignMixin
 
         kh: SignMixin = self._key_handle
-        spec = self._select_pkcs11_signing_params(digest_algorithm)
+        spec = self._select_pkcs11_signing_params(
+            digest_algorithm, sign_kwargs=self.sign_kwargs(data)
+        )
 
         if spec.pre_sign_transform is not None:
             data = spec.pre_sign_transform(data)
@@ -599,6 +619,87 @@ class PKCS11Signer(Signer):
     def _load_other_certs(self) -> Set[x509.Certificate]:
         return set(self.__pull())
 
+    def bulk_cert_fetch_params(self) -> Dict[Attribute, Any]:
+        """
+        Search parameters to bulk fetch certificates from the token.
+
+        .. note:: Can be overridden in subclasses.
+        """
+        return dict(self.default_cert_query_params)
+
+    def single_cert_fetch_params(
+        self,
+        label: Optional[str],
+        cert_id: Optional[bytes],
+    ) -> Dict[Attribute, Any]:
+        """
+        Search parameters to fetch a single certificate from the token.
+
+        .. note:: Can be overridden in subclasses.
+        """
+
+        query_params: Dict[Attribute, Any] = dict(
+            self.default_cert_query_params
+        )
+        if label is not None:
+            query_params[Attribute.LABEL] = label
+        if cert_id is not None:
+            query_params[Attribute.ID] = cert_id
+        return query_params
+
+    def signing_key_fetch_params(self) -> Dict[Attribute, Any]:
+        """
+        Search parameters to fetch the signing key from the token.
+
+        .. note:: Can be overridden in subclasses.
+        """
+        query_params: Dict[Attribute, Any] = dict(self.default_key_query_params)
+        if self.key_label is not None:
+            query_params[Attribute.LABEL] = self.key_label
+        if self.key_id is not None:
+            query_params[Attribute.ID] = self.key_id
+
+        return query_params
+
+    def _pull_single_cert(
+        self,
+        label: Optional[str],
+        cert_id: Optional[bytes],
+    ):
+        query_params = self.single_cert_fetch_params(label, cert_id)
+        q = self.pkcs11_session.get_objects(query_params)
+
+        # need to run through the full iterator to make sure the operation
+        # terminates
+        results = list(q)
+        if len(results) == 1:
+            cert_obj = results[0]
+            return x509.Certificate.load(cert_obj[Attribute.VALUE])
+        else:
+            err = _format_pull_err_msg(
+                "certificate",
+                no_results=not results,
+                label=label,
+                id_value=cert_id,
+            )
+            raise PKCS11Error(err)
+
+    def _pull_signing_key_handle(self):
+        query_params = self.signing_key_fetch_params()
+        q = self.pkcs11_session.get_objects(query_params)
+
+        results = list(q)
+        if len(results) == 1:
+            return results[0]
+        else:
+            err = _format_pull_err_msg(
+                "private key",
+                no_results=not results,
+                label=self.key_label,
+                id_value=self.key_id,
+            )
+            raise PKCS11Error(err)
+
     def __pull(self):
         other_cert_labels = self.other_certs
         if other_cert_labels is not None and len(other_cert_labels) == 0:
@@ -608,9 +709,7 @@ class PKCS11Signer(Signer):
             return
         if other_cert_labels is None or self.bulk_fetch:
             # first, query all certs
-            q = self.pkcs11_session.get_objects(
-                {Attribute.CLASS: ObjectClass.CERTIFICATE}
-            )
+            q = self.pkcs11_session.get_objects(self.bulk_cert_fetch_params())
             logger.debug("Pulling all certificates from PKCS#11 token...")
             for cert_obj in q:
                 label = cert_obj[Attribute.LABEL]
@@ -632,7 +731,7 @@ class PKCS11Signer(Signer):
                     f"PKCS#11 token..."
                 )
                 logger.debug(msg)  # lgtm
-                yield _pull_cert(self.pkcs11_session, label)
+                yield self._pull_single_cert(label=label, cert_id=None)
 
     async def ensure_objects_loaded(self):
         """
@@ -670,13 +769,11 @@ class PKCS11Signer(Signer):
 
         self._init_cert_registry()
         if self._signing_cert is None:
-            self._signing_cert = _pull_cert(
-                self.pkcs11_session, label=self.cert_label, cert_id=self.cert_id
+            self._signing_cert = self._pull_single_cert(
+                label=self.cert_label, cert_id=self.cert_id
             )
 
-        kh = self.pkcs11_session.get_key(
-            ObjectClass.PRIVATE_KEY, label=self.key_label, id=self.key_id
-        )
+        kh = self._pull_signing_key_handle()
         self._key_handle = kh
 
         self._loaded = True
@@ -718,7 +815,7 @@ class PKCS11SigningContext:
             raise SigningError(
                 f"PKCS#11 error while opening session to {config.module_path}: [{type(ex).__name__}] {ex}"
             ) from ex
-        return PKCS11Signer(
+        signer = PKCS11Signer(
             session,
             config.cert_label,
             ca_chain=config.other_certs,
@@ -732,6 +829,9 @@ class PKCS11SigningContext:
             signing_cert=config.signing_certificate,
             signature_mechanism=config.signature_mechanism,
         )
+        if config.only_resident_certs:
+            signer.default_cert_query_params[Attribute.TOKEN] = True
+        return signer
 
     def __enter__(self):
         return self._instantiate()
